# time tokei --output json > ~/tmp/tmp.json
# cat ~/tmp/tmp.json | python -m json.tool

import os
import json
import subprocess

SLOC_CMD = 'tokei'

required_cmds = {SLOC_CMD: 'install from https://github.com/XAMPPRocky/tokei/releases (put on path, chmod a+x)'}

_DEFAULT_IGNORED_LANGS = 'svg,ini,json,text'
IGNORED_LANGS = set([l.strip().lower() for l in
                     os.getenv('APATITE_SLOC_IGNORE_LANGS', _DEFAULT_IGNORED_LANGS).split(',')])
_DEFAULT_VERBOSE_LANGS = 'python,c,cpp'
VERBOSE_LANGS = set([l.strip().lower() for l in
                     os.getenv('APATITE_SLOC_VERBOSE_LANGS', _DEFAULT_VERBOSE_LANGS).split(',')])
_DEFAULT_IGNORED_FILES = 'translations,Language,contrib,vendor'  # freecad has a bunch of .ts files
IGNORED_FILES = set([l.strip() for l in
                     os.getenv('APATITE_SLOC_IGNORE_FILES', _DEFAULT_IGNORED_FILES).split(',')])

MERGE_HEADERS = not os.getenv('APATITE_NO_MERGE_HEADERS', '')

def run_cap(args, **kw):
    kw['stdout'] = subprocess.PIPE
    kw['stderr'] = subprocess.PIPE
    kw.setdefault('encoding', 'utf8')
    # slightly worried about ignoring utf8 decode errors on git
    # output, bc the git docs say they recode to utf8 (this came up bc
    # unicode characters in committer names)
    # https://git-scm.com/docs/git-show/1.8.2.2
    kw.setdefault('errors', 'replace')

    return subprocess.run(args, **kw)


def _dir_file_count(path):
    dir_count, file_count = 0, 0
    for root, dirs, files in os.walk(path):
        dir_count += len([d for d in dirs if not d.startswith('.')])
        file_count += len(files)
    return dir_count, file_count


def _merge_headers(d, lang):
    lang_code = '%s_code' % lang
    header_key = '%sheader' % lang
    header_code_key = '%sheader_code' % lang
    if not d.get(header_key):
        return
    try:
        d[lang] += d.get(header_key, 0)
        d[lang_code] += d.get(header_code_key, 0)
    except KeyError:
        d[lang] = d.get(header_key, 0)
        d[lang_code] = d.get(header_code_key, 0)

    d.pop(header_key, None)
    d.pop(header_code_key, None)
    return


def collect(plist, project, repo_dir):
    cmd = [SLOC_CMD, '--output', 'json']
    for file_part in IGNORED_FILES:
        cmd.extend(['-e', file_part])
    proc_res = run_cap(cmd, cwd=repo_dir)

    data = json.loads(proc_res.stdout)
    data = data['inner']
    ret = {}
    for lang, stats in data.items():
        if not stats['lines'] or stats['inaccurate']:
            continue
        lang = lang.lower()
        if lang in IGNORED_LANGS:
            continue
        if stats['lines'] < 50 and lang not in VERBOSE_LANGS:
            continue  # more noise reduction
        ret[lang] = stats['lines']
        if lang in VERBOSE_LANGS:
            ret[lang + '_code'] = stats['code']
            ret[lang + '_comments'] = stats['comments']
            ret[lang + '_files'] = len(stats['stats'])

    if MERGE_HEADERS:
        _merge_headers(ret, 'c')
        _merge_headers(ret, 'cpp')

    sorted_stats = sorted([(k, v) for k, v in ret.items() if '_' not in k],
                          reverse=True, key=lambda x: x[1])
    ret.update({'TOTAL_%s' % key: sum([s[key] for s in data.values()])
                for key in ['blanks', 'code', 'comments', 'lines']})

    ratio_map = {}
    for lang, count in sorted_stats:
        ratio = round(count / ret['TOTAL_lines'], 2)
        if not ratio or (ratio < 0.1 and lang not in VERBOSE_LANGS):
            continue
        ratio_map[lang.lower()] = ratio
    ratio_map['OTHER'] = round(1.0 - sum(ratio_map.values()), 2) if sorted_stats else 1.0

    ret['RATIO'] = ratio_map
    ret['TOTAL_dirs'], ret['TOTAL_files'] = _dir_file_count(repo_dir)

    return ret




import subprocess

def run_cap(args, **kw):
    kw['stdout'] = subprocess.PIPE
    kw['stderr'] = subprocess.PIPE
    kw.setdefault('encoding', 'utf8')
    # slightly worried about ignoring utf8 decode errors on git
    # output, bc the git docs say they recode to utf8 (this came up bc
    # unicode characters in committer names)
    # https://git-scm.com/docs/git-show/1.8.2.2
    kw.setdefault('errors', 'replace')

    return subprocess.run(args, **kw)



import re
import json
import datetime
import subprocess
from collections import Counter

import glom

from apatite.utils import run_cap

DETECT_CMD = 'license-detector'

required_cmds = {DETECT_CMD:
                 'Download from: https://github.com/src-d/go-license-detector/releases,'
                 ' rename to "license-detector" (if need be), place it on your PATH,'
                 ' and make it executable.'}

# missing = "other"
LICENSE_MAP = {
    'AFL-3.0': 'AFL',
    'AGPL-1.0': 'AGPL',
    'AGPL-3.0': 'AGPL',
    'AGPL-3.0-only': 'AGPL',
    'AGPL-3.0-or-later': 'AGPL',
    'Apache-2.0': 'Apache-2.0',
    'BSD-2-Clause': 'BSD-2',
    'BSD-2-Clause-FreeBSD': 'BSD-2',
    'BSD-2-Clause-NetBSD': 'BSD-2',
    'BSD-3-Clause': 'BSD-3',
    'BSD-3-Clause-Clear': 'BSD-3',
    'BSD-3-Clause-No-Nuclear-License-2014': 'BSD-3',
    'BSD-4-Clause': 'BSD-3',  # sure it exists, but not here
    'BSD-Source-Code': 'BSD-3',
    'CDDL-1.0': 'CDDL',
    'CDDL-1.1': 'CDDL',
    'GPL-2.0': 'GPL v2',
    'GPL-2.0-only': 'GPL v2',
    'GPL-2.0-or-later': 'GPL v2',
    'GPL-3.0-only': 'GPL v3',
    'GPL-3.0-or-later': 'GPL v3',
    'ISC': 'ISC',
    'LGPL-2.0+': 'LGPL v2',
    'LGPL-2.0-only': 'LGPL v2',
    'LGPL-2.0-or-later': 'LGPL v2',
    'LGPL-2.1': 'LGPL v2',
    'LGPL-2.1-only': 'LGPL v2',
    'LGPL-2.1-or-later': 'LGPL v2',
    'LGPL-3.0-only': 'LGPL v3',
    'LGPL-3.0-or-later': 'LGPL v3',
    'MIT': 'MIT',
    'MIT-feh': 'MIT',
    'MPL-1.1': 'MPL',
    'MPL-2.0': 'MPL',
    'MPL-2.0-no-copyleft-exception': 'MPL',
    'Unlicense': 'Unlicense',
    'deprecated_AGPL-3.0': 'AGPL  ',
    'deprecated_GPL-2.0': 'GPL v2',
    'deprecated_GPL-2.0+': 'GPL v2',
    'deprecated_GPL-3.0': 'GPL v3',
    'deprecated_GPL-3.0+': 'GPL v3',
    'deprecated_LGPL-2.0': 'LGPL v2',
    'deprecated_LGPL-2.0+': 'LGPL v2',
    'deprecated_LGPL-2.1': 'LGPL v2',
    'deprecated_LGPL-2.1+': 'LGPL v2',
    'deprecated_LGPL-3.0': 'LGPL v3',
    'deprecated_LGPL-3.0+': 'LGPL v3'}


# OSL: True
GROUP_HEREDITARY_MAP = {'AFL': False,
                        'AGPL': True,
                        'Apache': True,
                        'BSD': False,
                        'CDDL': False,
                        'GPL': True,
                        'ISC': True,
                        'LGPL': True,
                        'MIT': False,
                        'MPL': True,
                        'Unlicense': True}


'''
Too many false positives of these:
 'CC-BY-3.0': 'CC',
 'CC-BY-4.0': 'CC',
 'CC0-1.0': 'CC0',
'''


def collect(plist, project, repo_dir):
    ret = {}
    proc_res = run_cap([DETECT_CMD, repo_dir, '--format', 'json'])
    output_json = json.loads(proc_res.stdout)
    possible_licenses = glom.glom(output_json, '0.matches', default=[])
    # sort and set into descending order
    possible_licenses = sorted(possible_licenses, key=lambda x: x['confidence'], reverse=True)[:3]
    norm_licenses = []
    for pl in possible_licenses:
        if pl['confidence'] < 0.9:
            continue
        elif pl['license'] not in LICENSE_MAP:
            continue
        norm_licenses.append((LICENSE_MAP[pl['license']], round(pl['confidence'], 3)))

    if not norm_licenses or len(norm_licenses) > 3:
        ret['license'] = 'Other'  # not enough consensus on a known license
    else:
        sorted(norm_licenses, key=lambda x: x[1], reverse=True)
        if len(norm_licenses) < 3:
            ret['license'] = norm_licenses[0][0]
        else:
            most_common = Counter([x[0] for x in norm_licenses]).most_common(1)[0][0]
            ret['license'] = most_common

    group = re.split('\W+', ret['license'])[0]
    ret['license_group'] = group
    ret['hereditary'] = GROUP_HEREDITARY_MAP.get(group)

    return ret





def collect(plist, project, repo_dir):
    topics = _get_topics(plist, project)
    return {'primary': topics[0],
            'primary_fq': topics[1],
            'secondary': topics[2],
            'secondary_fq': topics[3]}

def _get_topics(plist, project):
    tags = [t for t in project._tags if t in plist.tag_registry]  # TODO: why is _tags underscored again?

    topic_tags = [t for t in tags if plist.tag_registry[t].tag_type == 'topic']

    primary = None
    secondary = None
    if not topic_tags:
        print(project.name)
    elif len(topic_tags) >= 1:
        primary = topic_tags[0]
    if len(topic_tags) >= 2:
        secondary = topic_tags[1]

    return (primary, plist.tag_registry[primary].fq_tag if primary else None,
            secondary, plist.tag_registry[secondary].fq_tag if secondary else None)




from boltons.iterutils import soft_sorted

# house = u"\u2302"
BULLET = '1.'
INDENT = ' ' * 4

# sort of document the expected ones, even when they match the
# .title() pattern
_URL_LABEL_MAP = {'wp': 'WP',
                  'gh': 'gh',
                  'home': 'Home',
                  'repo': 'Repo',
                  'docs': 'Docs',
                  'pypi': 'PyPI'}

_URL_ORDER = ['repo', 'home', 'wp', 'docs']


def _format_url_name(name):
    return _URL_LABEL_MAP.get(name, name.title())

def get_url_list(project):
    urls = [u for u in project.urls if u[0] != 'clone']  # clone_urls aren't for display
    return [(_format_url_name(name), url) for name, url in
            soft_sorted(urls, key=lambda x: x[0], first=_URL_ORDER[:-1], last=_URL_ORDER[-1:])]


def format_category(project_map, tag_entry):
    lines = []
    append = lines.append

    def _format_tag(project_map, tag_entry, level=2):
        append('%s <a id="tag-%s" href="#tag-%s">%s</a>' %
               ('#' * level, tag_entry.fq_tag or tag_entry.tag, tag_entry.fq_tag or tag_entry.tag, tag_entry.title))
        append('')
        if tag_entry.desc:
            append(tag_entry.desc)
            append('')
        if tag_entry.subtags:
            append('')
            for subtag_entry in tag_entry.subtags:
                _format_tag(project_map, subtag_entry, level=level + 1)
            append('%s <a id="tag-%s-other" href="#tag-%s-other">Other %s projects</a>' %
                   ('#' * (level + 1), tag_entry.tag, tag_entry.tag, tag_entry.title))

        for project in project_map[tag_entry]:
            tmpl = '  {bullet} **{name}** - ({links}) {desc}'
            links = ', '.join(['[%s](%s)' % item for item in get_url_list(project)])

            line = tmpl.format(bullet=BULLET, name=project.name, links=links, desc=project.desc)
            if len(project._tags) > 1:
                other_tags = [t for t in project._tags if t != tag_entry.tag]
                line += ' `(%s)`' % ', '.join(other_tags)
            lines.append(line)

        append('')
        return '\n'.join(lines)

    return _format_tag(project_map, tag_entry)


def format_tag_toc(project_map):
    lines = []

    def _format_tag_toc(tag_entries, path=()):
        for te in tag_entries:
            if te.tag_path != path:
                continue
            entry_count = len(project_map[te])
            if te.subtags:
                entry_count = len(project_map[te]) + len(set.union(*[set(project_map[st]) for st in te.subtags]))
            link_text = '<a href="#tag-%s">%s</a> *(%s)*' % (te.fq_tag or te.tag, te.title, entry_count)
            lines.append((INDENT * len(te.tag_path)) + BULLET + ' ' + link_text)
            if te.subtags:
                _format_tag_toc(te.subtags, path=path + (te.tag,))
                if len(project_map[te]):
                    link_text = ('<a href="#tag-%s-other">Other %s projects</a> *(%s)*'
                                 % (te.fq_tag or te.tag, te.title, len(project_map[te])))
                    lines.append((INDENT * (len(te.tag_path) + 1)) + BULLET + ' ' + link_text)
        return

    _format_tag_toc(project_map.keys())

    return '\n'.join(lines)


def format_all_categories(project_map):
    parts = []
    for tag_entry in project_map:
        if tag_entry.tag_path:
            continue
        if not project_map[tag_entry] and not tag_entry.subtags:
            continue  # TODO: some message, inviting additions
        text = format_category(project_map, tag_entry)
        parts.append(text)

    return '\n'.join(parts)



import glom
import json
import os
import urllib.request


API_KEY_ENV_VAR = 'GH_API_KEY'

required_cmds = {}
required_env_vars = {API_KEY_ENV_VAR: 'The GitHub API only allows 50 requests per hour for unauthenticated clients.'
                     '\nTo unlock 5000/hour, sent an env var with a token generated here: https://github.com/settings/tokens'
                     '\nIf this is your first time, it might help to follow the directions here:'
                     ' https://help.github.com/en/articles/creating-a-personal-access-token-for-the-command-line.'
                     '\nNote that no special permissions are needed, so leave all/most of those checkboxes unchecked.'}


def collect(plist, project, repo_dir):

    host = project.repo_url.host
    if host != 'github.com':
        return {}
    GH_API_KEY = os.getenv(API_KEY_ENV_VAR)
    api_url = 'https://api.github.com/repos/%s/%s' % (project.author,
                                                      project.repo_name)
    request = urllib.request.Request(api_url)
    request.add_header('Authorization', 'token %s' % GH_API_KEY)
    response = urllib.request.urlopen(request)
    api_json = json.loads(response.read())
    glom_spec = {
        'size': 'size',
        'stars': 'stargazers_count',
        'watchers': 'subscribers_count',
        'forks': 'forks',
        'open_issues': 'open_issues',
        'license': glom.Coalesce('license.name', default=None),
        'has_wiki': 'has_wiki',
    }
    return glom.glom(api_json, glom_spec)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
import io
import os
import copy
import datetime

import attr
import schema
from glom import glom, T, Coalesce
from ruamel.yaml import YAML, round_trip_load
from ruamel.yaml.comments import CommentedMap
from boltons.dictutils import OMD
from boltons.fileutils import iter_find_files, atomic_save
from boltons.iterutils import soft_sorted, redundant, unique
from boltons.timeutils import isoparse
from boltons.strutils import slugify
from hyperlink import parse as url_parse

TOOLS_PATH = os.path.dirname(os.path.abspath(__file__))
TEMPLATES_PATH = os.path.dirname(os.path.abspath(__file__)) + '/templates/'

# house = u"\u2302"
BULLET = '1.'
INDENT = ' ' * 4

PLIST_PREFACE_COMMENT = """\
Format of this file:

 - First we have the "tagsonomy", a tree of tags used to categorize the projects.
 - After that is "projects", a list of awesome projects

Each project has the following format:

 - name: Project Name
   repo_url: github or bitbucket or other web link to code
   wp_url: Wikipedia URL if there is one
   docs_url: URL to docs
   home_url: Home page if not one of the above
   tags: ["", ""]  # see full taxonomy elsewhere in this file
   desc: Handy project designed for handiness

Only the name, repo_url, tags, and description are required.

In the description, avoid references to Python, free/open-source, and
the app name itself, since those are already present/implied by being
on the list.  """


def parse_valid_url(url, schemes=None):
    url_obj = url_parse(url)
    url_obj = url_obj.normalize()

    assert url_obj.scheme
    if schemes is not None:
        assert url_obj.scheme in schemes
    assert url_obj.host

    return url_obj


def parse_valid_web_url(url):
    return parse_valid_url(url, schemes=('http', 'https'))


_PROJECT_SCHEMA = schema.Schema(
    {'name': str,
     schema.Optional(schema.Regex('\w+_url')): parse_valid_url,
     'repo_url': parse_valid_web_url,
     schema.Optional('date_added'): datetime.datetime,
     'desc': str,
     'tags': tuple},
    ignore_extra_keys=False)


class ApatiteError(Exception):
    pass


class DuplicateProjectError(ApatiteError):
    pass


class ProjectValidationError(ApatiteError):
    @classmethod
    def from_pdict_error(cls, pdict, error):
        msg = '%s while validating project' % error

        name = pdict.get('name')
        if name is not None:
            msg += ' %r' % name

        line = glom(pdict, T.lc.line, default=None)
        if line is not None:
            msg += ' on line %r' % line

        if name is None and line is None:
            msg += ' %r' % pdict

        ret = cls(msg)
        ret.pdict = pdict
        ret.error = error
        return ret


class ProjectListError(ApatiteError):
    @classmethod
    def from_errors(cls, errors):
        msg = 'encountered %s errors in the project list:' % len(errors)
        msg += '\n'.join([''] + ['  - ' + repr(e) for e in errors])
        ret = cls(msg)
        ret.errors = errors
        return ret


def validate_project_dict(pdict):
    try:
        return _PROJECT_SCHEMA.validate(pdict)
    except schema.SchemaError as se:
        raise ProjectValidationError.from_pdict_error(pdict, se)


def to_yaml(obj):
    sio = io.StringIO()
    yaml = YAML()
    yaml.width = 2000
    yaml.indent(mapping=2, sequence=4, offset=2)
    yaml.dump(obj, sio)
    return sio.getvalue()


@attr.s(frozen=True)
class TagEntry(object):
    tag = attr.ib()
    tag_type = attr.ib()
    title = attr.ib()
    desc = attr.ib(default='')
    subtags = attr.ib(default=(), repr=False)
    tag_path = attr.ib(default=(), repr=False)
    fq_tag = attr.ib(default=None)

    @property
    def is_fq(self):
        return self.tag == self.fq_tag


@attr.s(frozen=True)
class Project(object):
    name = attr.ib()
    desc = attr.ib(default='')
    _tags = attr.ib(default=())
    urls = attr.ib(default=())
    date_added = attr.ib(default=attr.Factory(lambda: datetime.datetime.now().replace(second=0, microsecond=0)))
    _orig_data = attr.ib(default=None, repr=False, cmp=False)

    @property
    def name_slug(self):
        return slugify(self.name)

    @property
    def repo_url(self):
        for name, url in self.urls:
            if name == 'repo':
                return url
        return None

    @property
    def author(self):
        if self.repo_url.host != 'github.com':
            return None
        return self.repo_url.path[0]

    @property
    def repo_name(self):
        if self.repo_url.host != 'github.com':
            return None
        return self.repo_url.path[1]

    @property
    def clone_info(self):
        for name, url in self.urls:
            if name == 'clone':
                vcs, _, scheme = url.scheme.rpartition('+')
                clone_url = url.replace(scheme=scheme)
                return (vcs, clone_url)

        repo_url = self.repo_url

        if repo_url.host == 'github.com' or 'gitlab.' in repo_url.host:  # covers gitlab.gnome.org, too
            return ('git', repo_url.replace(path=repo_url.path[:-1] + (repo_url.path[-1] + '.git',)))
        elif repo_url.path[-1].endswith('.git'):
            return ('git', repo_url)
        elif repo_url.host == 'bitbucket.org':
            return ('hg', repo_url.replace(path=(repo_url.path[0], repo_url.path[1])))
        elif repo_url.host == 'code.launchpad.net' or repo_url.host == 'launchpad.net':
            return ('bzr', 'lp:' + repo_url.path[0])
        return (None, None)

    @classmethod
    def from_dict(cls, d):
        validate_project_dict(d)
        kwargs = dict(d)
        tags = tuple([slugify(t) for t in kwargs.get('tags', ())])
        dupe_groups = redundant(tags, groups=True)
        if dupe_groups:
            raise ProjectValidationError('duplicate tags in project %r: %r'
                                         % (kwargs['name'], dupe_groups))
        kwargs['tags'] = tags
        cur_urls = ()
        for k in list(kwargs):
            if not k.endswith('_url'):
                continue
            val = kwargs.pop(k)
            val = parse_valid_url(val)
            cur_urls += ((k[:-4], val),)
            cur_urls = soft_sorted(cur_urls, first=['repo', 'home'], last=['wp'], key=lambda x: x[0])
            kwargs['urls'] = tuple(cur_urls)
        kwargs['orig_data'] = d
        return cls(**kwargs)

    def to_dict(self):
        # deepcopy necessary to maintain comments
        ret = copy.deepcopy(self._orig_data) if self._orig_data else CommentedMap()
        ret['name'] = self.name
        for i, (url_type, url) in enumerate(self.urls):
            key = url_type + '_url'
            ret.pop(key, None)
            ret.insert(1 + i, key, str(url))
        if self.date_added:
            ret['date_added'] = self.date_added
        ret['desc'] = self.desc
        ret.move_to_end('desc')
        ret['tags'] = self._tags
        ret.move_to_end('tags')
        return ret


def _unwrap_dict(d):
    if not len(d) == 1:
        raise ValueError('expected single-member dict')
    return list(d.items())[0]


class ProjectList(object):
    def __init__(self, project_list, tagsonomy):
        self.project_list = []
        self.tagsonomy = tagsonomy
        self.tag_registry = OMD()

        for tag_group in ('topic', 'platform'):  # TODO: framework, license
            for tag in self.tagsonomy[tag_group]:
                self.register_tag(tag_group, tag)

        errors = []
        for project in project_list:
            new_tags = tuple(soft_sorted(project.get('tags', []), first=self.tag_registry.keys()))
            project['tags'] = new_tags
            try:
                project_obj = Project.from_dict(project)
            except ApatiteError as ae:
                errors.append(ae)
                continue
            self.project_list.append(project_obj)

        dupe_groups = redundant(self.project_list,
                                key=lambda p: slugify(p.name),
                                groups=True)
        dupe_groups += redundant(self.project_list,
                                 key=lambda p: p.repo_url,
                                 groups=True)
        dupe_groups = unique([tuple(dg) for dg in dupe_groups])
        for group in dupe_groups:
            dpe = DuplicateProjectError('ambiguous or duplicate projects: %r' %
                                        [p.name for p in group])
            errors.append(dpe)

        if errors:
            raise ProjectListError.from_errors(errors)
        return

    @classmethod
    def from_path(cls, path):
        data = round_trip_load(open(path, encoding='utf-8'))
        return cls.from_dict(data)

    @classmethod
    def from_dict(cls, data):
        return cls(project_list=data['projects'],
                   tagsonomy=data['tagsonomy'])

    def to_dict(self):
        ret = CommentedMap()
        ret.yaml_set_start_comment('\n' + PLIST_PREFACE_COMMENT + '\n\n')
        ret['tagsonomy'] = self.tagsonomy
        plist = []
        seen_topics = set()
        for p in self.project_list:
            cur_pdict = p.to_dict()
            plist.append(cur_pdict)

            # now, determine whether to emit a comment
            topic_tags = [t for t in p._tags
                          if t in self.tag_registry
                          and self.tag_registry[t].tag_type == 'topic']
            if not topic_tags:
                continue
            first_topic = topic_tags[0]
            if first_topic in seen_topics:
                continue
            seen_topics.add(first_topic)
            cur_pdict.yaml_set_start_comment('\n' + first_topic.title() + '\n\n')

        ret['projects'] = plist
        return ret

    def to_yaml(self):
        return to_yaml(self.to_dict())

    def get_project(self, name):
        name_slug = slugify(name)
        for proj in self.project_list:
            if proj.name_slug == name_slug:
                return proj
        raise LookupError('no such project: %r' % name)

    def update_projects(self, projects):
        to_update = OMD([(slugify(p.name), p) for p in projects])
        new_list = []
        for proj in self.project_list:
            if proj.name_slug not in to_update:
                new_list.append(proj)
                continue
            new_list.append(to_update.pop(proj.name_slug))
        new_list.extend(to_update.values())
        self.project_list = new_list

    def normalize(self):
        # sort project list by first topic tag and name (lexi).
        tag_list = list(self.tag_registry.keys())
        def plist_sort_key(project):
            topic_tags = [t for t in project._tags
                          if t in self.tag_registry
                          and self.tag_registry[t].tag_type == 'topic']
            first_topic = topic_tags[0] if topic_tags else 'misc'  # TODO: might change to uncategorized in future
            first_topic_idx = tag_list.index(first_topic)
            return (first_topic_idx, project.name.lower())

        project_list = []
        for project in sorted(self.project_list, key=plist_sort_key):
            if not project.desc[-1:] in ').!':
                project = attr.evolve(project, desc=project.desc + '.')

            cleaned_urls = []
            for url_name, url in project.urls:
                # strip off trailing slashes from all urls
                new_path = tuple([segm for segm in url.path if segm != ''])
                clean_url = url.replace(path=new_path).normalize()
                cleaned_urls.append((url_name, clean_url))
            project = attr.evolve(project, urls=tuple(cleaned_urls))

            project_list.append(project)

        self.project_list = project_list
        return

    def register_tag(self, tag_type, tag_entry, tag_path=()):
        if isinstance(tag_entry, str):
            tag, tag_entry = tag_entry, {}
        else:
            tag, tag_entry = _unwrap_dict(tag_entry)
            tag_entry = dict(tag_entry)
        tag_entry['tag'] = tag
        tag_entry['tag_type'] = tag_type

        if not tag_entry.get('title'):
            tag_entry["title"] = tag.replace('_', ' ').title()

        subtags = []
        for subtag_entry in tag_entry.pop('subtags', []):
            st = self.register_tag(tag_type, subtag_entry,
                                   tag_path=(tag,) if not tag_path else tag_path + (tag,))
            subtags.append(st)
        tag_entry['subtags'] = tuple(subtags)
        tag_entry['fq_tag'] = '.'.join(tag_path + (tag,))
        if not tag_path:
            ret = TagEntry(**tag_entry)
        else:
            ret = TagEntry(tag_path=tag_path, **tag_entry)

        self.tag_registry[tag] = ret
        return ret

    def get_projects_by_type(self, type_name):
        ret = OMD()
        for tag, tag_entry in self.tag_registry.items():
            if tag_entry.tag_type != type_name:
                continue
            ret[tag_entry] = []
            for project in self.project_list:
                if tag in project._tags:
                    ret[tag_entry].append(project)
            ret[tag_entry].sort(key=lambda x: x.name.lower())
        return ret


@attr.s
class ResultFile(object):
    entries = attr.ib()



# -*- coding: utf-8 -*-

import os
import imp
from setuptools import setup, find_packages

__author__ = "Mahmoud Hashemi"
__contact__ = "mahmoud@hatnote.com"
__license__ = 'GPLv3'
__url__ = 'https://github.com/mahmoud/apatite'

CUR_PATH = os.path.abspath(os.path.dirname(__file__))
_version_mod_path = os.path.join(CUR_PATH, 'apatite', '_version.py')
_version_mod = imp.load_source('_version', _version_mod_path)
__version__ = _version_mod.__version__

setup(
    name="apatite",
    description="CLI for awesomer Awesome™ list management and analysis. Originally designed for managing https://github.com/mahmoud/awesome-python-applications",
    author=__author__,
    author_email=__contact__,
    url=__url__,
    license=__license__,
    platforms='any',
    version=__version__,
    long_description=__doc__,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    entry_points={'console_scripts': ['apatite = apatite.__main__:main']},
    install_requires=['attrs',
                      'boltons',
                      'face',
                      'ruamel.yaml',
                      'glom',
                      'hyperlink',
                      'schema',
                      'tqdm']
)

"""
Release process:

* tox  # TODO
* git commit (if applicable)
* Remove dev suffix from apatite/_version.py version
* git commit -a -m "bump version for vX.Y.Z release"
* python setup.py sdist bdist_wheel upload
* git tag -a vX.Y.Z -m "brief summary"
* write CHANGELOG
* git commit
* bump apatite/_version.py version onto n+1 dev
* git commit
* git push

Versions are of the format YY.MINOR.MICRO, see calver.org for more details.
"""



# -*- coding: utf-8 -*-

# TODO: check - README should consist of nothing but words in
# templates and yaml. any other words and it's a sign the README/other
# docs have been manually edited.

import os
import csv
import imp
import sys
import json
import time
import shutil
import datetime
import subprocess
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor, as_completed

import attr
import glom
from tqdm import tqdm
from ashes import AshesEnv
from face import Command, Flag, face_middleware, ListParam
from boltons.fileutils import iter_find_files, atomic_save, mkdir_p, iter_find_files
from boltons.timeutils import isoparse, parse_timedelta
from boltons.jsonutils import JSONLIterator
from boltons.setutils import IndexedSet
from boltons.iterutils import remap, first

from .dal import ProjectList
from .formatting import format_tag_toc, format_all_categories, get_url_list
from ._version import __version__
from .utils import run_cap

_ANSI_FORE_RED = '\x1b[31m'
_ANSI_FORE_GREEN = '\x1b[32m'
_ANSI_RESET_ALL = '\x1b[0m'

CUR_PATH = os.path.dirname(os.path.abspath(__file__))
METRICS_PATH = os.path.join(CUR_PATH, 'metrics')


def print_err(*a, **kw):
    kw['file'] = sys.stderr
    return print(*a, **kw)


def _get_colorized_lines(lines):
    ret = []
    colors = {'-': _ANSI_FORE_RED, '+': _ANSI_FORE_GREEN}
    for line in lines:
        if line[0] in colors:
            line = colors[line[0]] + line + _ANSI_RESET_ALL
        ret.append(line)
    return ret


class APAError(Exception):
    pass

class APACLIError(APAError):
    "Raised when there's an error from user input or other CLI interaction"
    def __init__(self, msg=None, exit_code=1):
        self.msg = msg
        self.exit_code = exit_code
        if msg:
            super(APACLIError, self).__init__(msg)
        return


def _date_param(text):
    text = text.strip()
    if text.startswith('-'):
        td = parse_timedelta(text)
        dt = datetime.datetime.utcnow() + td  # adding a negative
        return dt
    dt = isoparse(text)
    return dt


def main(argv=None):
    """\
    automation and analytics for curated lists of awesome software.

    Normal analysis workflow:

      * apatite pull-repos  (can take 3-4 hours, 25GB on the full APA, use --targets to limit)
      * apatite collect-metrics
      * apatite export-metrics
      * apatite analyze  # TODO
    """
    cmd = Command(name='apatite', func=None, doc=main.__doc__)  # func=None means output help

    # add flags
    cmd.add('--file', missing='projects.yaml',
            doc='path to the project listing YAML file')
    cmd.add('--confirm', parse_as=True,
            doc='show diff and prompt for confirmation before modifying the file')
    cmd.add('--non-interactive', parse_as=True,
            doc='disable falling back to interactive authentication, useful for automation')
    cmd.add('--targets', parse_as=ListParam(str), missing=[], doc='specific target projects')
    cmd.add('--metrics', parse_as=ListParam(str), missing=[], doc='specific metrics to collect')
    cmd.add('--dry-run', parse_as=True, doc='do not save results')
    two_weeks_ago = _date_param('-2w')
    cmd.add('--earliest', parse_as=_date_param, missing=two_weeks_ago,
            doc=('minimum datetime value to accept (isodate or negative timedelta).'
                 ' defaults to two weeks ago (-2w)'))
    cmd.add('--no-progress', parse_as=True)

    # add middlewares, outermost first ("first added, first called")
    cmd.add(mw_exit_handler)
    cmd.add(mw_ensure_project_listing)
    cmd.add(mw_ensure_work_dir)

    # add subcommands
    cmd.add(render)
    cmd.add(normalize)
    cmd.add(pull_repos)
    cmd.add(collect_metrics)
    cmd.add(show_recent_metrics)
    cmd.add(export_metrics)
    cmd.add(show_exportable_metrics)
    cmd.add(set_repo_added_dates)
    cmd.add(console)
    cmd.add(print_version, name='version')

    cmd.prepare()  # an optional check on all subcommands, not just the one being executed

    try:
        cmd.run(argv=argv)  # exit behavior is handled by mw_exit_handler
    except Exception:
        if os.getenv('APATITE_DEBUG'):
            import pdb;pdb.post_mortem()
        raise

    return


def console(plist, pdir):
    "use pdb to explore plist and pdir"
    import pdb;pdb.set_trace()
    return


def _get_commit_dt(text):
    return isoparse(text.rsplit(':', 1)[0]).replace(second=0, microsecond=0)


def set_repo_added_dates(pfile, plist, targets, dry_run):
    project_list = plist.project_list
    if targets:
        project_list = [proj for proj in project_list if (proj.name in targets or proj.name_slug in targets)]
    results = []
    updated_proj_list = []
    for project in tqdm(project_list):
        if project._orig_data.get('date_added'):
            continue
        # rstrip in case it was added in a denormalized form
        repo_url = str(project.repo_url).rstrip('/')
        res = run_cap(['git', 'log', '--date=iso-strict', '--pretty=format:"%h%x09%an%x09%ad%x09%s"',
                       '--reverse', '--source', '-S', repo_url, '--', pfile])
        first_line = first(res.stdout.splitlines())
        if not first_line:
            print('nothing for', repo_url)
            continue  # TODO
        parts = [p.strip() for p in first_line.strip('"').split('\t')]
        dt_text = parts[2]
        parsed_date = _get_commit_dt(dt_text)
        results.append((project.name, parsed_date))
        updated_proj_list.append(attr.evolve(project, date_added=parsed_date))
    results.sort(key=lambda x: x[1])
    plist.update_projects(updated_proj_list)
    if not dry_run and updated_project_list:
        normalize(pfile=pfile, plist=plist)
    return


def render(plist, pdir, pfile):
    "generate the list markdown from the yaml listing"
    normalize(pfile=pfile, plist=plist)
    topic_map = plist.get_projects_by_type('topic')
    topic_toc_text = format_tag_toc(topic_map)
    projects_by_topic = format_all_categories(topic_map)

    plat_map = plist.get_projects_by_type('platform')
    plat_toc_text = format_tag_toc(plat_map)
    projects_by_plat = format_all_categories(plat_map)

    context = {'TOPIC_TOC': topic_toc_text,
               'TOPIC_TEXT': projects_by_topic,
               'PLATFORM_TOC': plat_toc_text,
               'PLATFORM_TEXT': projects_by_plat,
               'TOTAL_COUNT': len(plist.project_list)}

    templates_path = pdir + '/templates/'
    if not os.path.isdir(templates_path):
        raise APACLIError('expected "templates" directory at %r' % templates_path)

    for filename in iter_find_files(templates_path, '*.tmpl.md'):
        tmpl_text = open(filename).read()
        target_filename = os.path.split(filename)[1].replace('.tmpl', '')
        output_text = tmpl_text.format(**context)
        with atomic_save(pdir + '/' + target_filename) as f:
            f.write(output_text.encode('utf8'))

    feed_tmpl_path = templates_path + '/atom.xml'
    if os.path.exists(feed_tmpl_path):
        def _stderr_log_func(level, name, message):
            import sys
            sys.stderr.write('%s - %s - %s\n' % (level.upper(), name, message))
            sys.stderr.flush()

        ashes_env = AshesEnv([templates_path], log_func=_stderr_log_func)
        proj_dict_list = []
        for proj in plist.project_list:
            cur = proj.to_dict()
            cur['name_slug'] = proj.name_slug
            cur['date_added_utc'] = proj.date_added.isoformat() + 'Z'
            cur['urls'] = get_url_list(proj)
            proj_dict_list.append(cur)
        cur_dt = datetime.datetime.utcnow().replace(microsecond=0).isoformat() + 'Z'
        res = ashes_env.render('atom.xml',
                               {'projects': sorted(proj_dict_list, key=lambda x: x['date_added'], reverse=True),
                                'last_generated_utc': cur_dt})
        with atomic_save(pdir + '/atom.xml') as f:
            f.write(res.encode('utf8'))


    return


def check(plist):
    """
    [x] Check for the minimum set of keys.
    [x] Check all URLs valid format
    [x] Check no duplicate names (case normalized)
    [x] Check no duplicate repo_urls
    [x] Check no duplicate tags within the same project tag list
    [ ] Check all tags are in the tagsonomy

    All this stuff should probably be done before rendering
    anyways. Might need a check_urls subcommand though.

    """
    return 0


def show_no_cat(cat_name):
    """
    Show projects which are missing tags for a certain category.

    (this may be best covered by just rendering these into the docs.
    """


def normalize(plist, pfile):
    """normalize project and tag order, checking for duplicates
    and format divergences, overwrites the yaml listing"""
    plist.normalize()
    new_yaml = plist.to_yaml()
    # say no to trailing whitespace
    new_yaml = '\n'.join([line.rstrip() for line in new_yaml.splitlines()])
    with atomic_save(pfile) as f:
        f.write(new_yaml.encode('utf8'))
    return


@attr.s
class ProcessResult(object):
    returncode = attr.ib()
    stderr = attr.ib()
    stdout = attr.ib()
    start_time = attr.ib()
    end_time = attr.ib()

    @property
    def duration(self):
        return self.end_time - self.start_time


"""what's going to grab the pid for cancellation?

problems with concurrent futures:

* Why does the future need to know about the executor? Why can't we
create a future and submit that to the executor?

* It's seeming like I have to map arguments to futures and processes
to arguments, and then I can associate futures to processes?

"""

def format_list(list_tmpl, *a, **kw):
    ret = []
    for text in list_tmpl:
        ret.append(text.format(*a, **kw))
    return ret


VCS_TMPLS = {'git': {'clone': ['{cmd}', 'clone', '{url}', '{target_dir}'],
                     'update': ['{cmd}', 'pull']},  # update is run from within target_dir
             'hg': {'clone': ['{cmd}', 'clone', '{url}', '{target_dir}'],
                    'update': ['{cmd}', 'pull']},
             'bzr': {'clone': ['{cmd}', 'branch', '{url}'],
                     'update': ['{cmd}', 'pull']}}
CUR_PROCS = []


def _pull_single_repo(proj, repo_dir, rm_cached=False):
    # TODO: turn rm_cached into a flag
    # TODO: shouldn't this be the callable submitted to the executor?
    vcs, url = proj.clone_info
    if url is None:
        print_err('project "%s" has unsupported vcs type for repo url: %r' % (proj.name, proj.repo_url))
        return  # TODO
    target_dir = repo_dir + proj.name_slug + '/'
    cwd = repo_dir
    mode = 'clone'
    if os.path.exists(target_dir):
        if not rm_cached:
            mode = 'update'
            cwd = target_dir
        else:
            # sanity check we're in the right place
            if os.path.exists(target_dir + '../.apatite_repo_dir'):
                shutil.rmtree(target_dir)
            else:
                raise Exception('non-apatite path blocking apatite clone: %r' % target_dir)

    try:
        cmd_tmpl = VCS_TMPLS[vcs][mode]
    except KeyError:
        raise Exception('unsupported operation %r with vcs %r' % (mode, vcs))
    cmd = format_list(cmd_tmpl, cmd=vcs, url=url, target_dir=target_dir)
    # avoid asking for github username and password, since closing stdin doesn't work

    cur_dt = datetime.datetime.utcnow()
    started = time.time()
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd, env={'GIT_TERMINAL_PROMPT': '0'})
    proc.project_name = proj.name_slug
    proc.target_dir = target_dir
    CUR_PROCS.append(proc)
    stdout, stderr = proc.communicate()
    CUR_PROCS.remove(proc)
    proc_res = ProcessResult(returncode=proc.returncode,
                             stdout=stdout.decode('utf8'),
                             stderr=stderr,
                             start_time=started,
                             end_time=datetime.datetime.utcnow()) # Bug? duratiion, start_time - end_time TypeError
    if proc_res.returncode != 0:
        print_err('%r exited with code %r, stderr:' % (proc.args, proc_res.returncode))
        print_err(proc_res.stderr)
    else:
        with atomic_save(target_dir + '/.apatite_last_pulled') as f:
            f.write(cur_dt.isoformat().partition('.')[0].encode('utf8'))

    return proc_res


def pull_repos(plist, targets, repo_dir, verbose=False):
    """
    clone or pull all projects. requires git, hg, and bzr to be installed for projects in APA
    """
    successes = []
    try:
        project_list = plist.project_list
        if targets:
            project_list = [proj for proj in project_list if (proj.name in targets or proj.name_slug in targets)]
        with ThreadPoolExecutor(max_workers=5) as executor, tqdm(total=len(project_list)) as progress:
            futs = []

            for proj in project_list:
                fut = executor.submit(_pull_single_repo, proj, repo_dir)
                fut.project = proj
                futs.append(fut)

            for fut in as_completed(futs):
                progress.update()
                progress.set_description(desc='(%s)' % ', '.join([fut.project.name_slug for fut in futs if fut.running()]))
                exc = fut.exception()
                if exc is not None:
                    print_err('%s got exception %r' % (fut.project.name_slug, exc))
                    continue
                proc_res = fut.result()
                if not proc_res:
                    continue  # TODO
                if proc_res.returncode == 0:
                    successes.append(fut.project)
                if verbose:
                    print(proc_res.stdout)
    finally:
        for proc in CUR_PROCS:
            if proc.returncode:
                continue
            proc.kill()
            # clean up incomplete paths
            if os.path.exists(proc.target_dir):
                shutil.rmtree(proc.target_dir)
    print()
    if successes:
        print('successfully pulled %s project(s): %s' % (len(successes), ', '.join([p.name_slug for p in successes])))
    else:
        print('failed to pull any projects')
    return


class ProjectProcessor(object):
    # TODO
    def __init__(self, func, args):
        self.futs = []
        self.func = func
        self.args = args

    @property
    def results(self):
        return [f.result() for f in self.futs]

    def process(self):
        # TODO: basically the generic parts of pull_repos
        pass


def _get_project_repo_info_map(project_list, repo_dir):
    ret = OrderedDict()

    for project in project_list:
        target_repo_dir = os.path.join(repo_dir, project.name_slug)
        if not os.path.isdir(target_repo_dir):
            print_err('project %s repo directory not found at: %r' % (project.name, target_repo_dir))
            continue
        pull_date_path = target_repo_dir + '/.apatite_last_pulled'
        with open(pull_date_path, 'r') as f:
            last_pulled_bytes = f.read()
            try:
                last_pulled = isoparse(last_pulled_bytes)
            except (TypeError, ValueError):
                print_err('project %s had unreadable pull date at: %r' % (project.name, pull_date_path))
                continue
        ret[project] = (target_repo_dir, last_pulled)

    return ret


def _get_all_metric_mods(check_reqs=True):
    ret = []
    for metric_path in iter_find_files(METRICS_PATH, '*.py', ignored='__init__.py'):
        mod_name = os.path.splitext(os.path.split(metric_path)[-1])[0]
        metric_mod = imp.load_source(mod_name, metric_path)
        if not callable(getattr(metric_mod, 'collect', None)):
            print_err('skipping non-metric module at %r' % metric_path)
            continue
        if not check_reqs:
            ret.append(metric_mod)
            continue
        missing_env_vars = _check_required_env_vars(metric_mod)
        missing_cmds = _check_required_cmds(metric_mod)
        if missing_cmds:
            print_err('omitting metric "%s" due to missing commands: %s (see installation instructions above)'
                      % (metric_mod.__name__, ', '.join(missing_cmds)))
        elif missing_env_vars:
            print_err('omitting metric "%s" due to missing ENV variables: %s'
                      % (metric_mod.__name__, ', '.join(missing_env_vars)))
        else:
            ret.append(metric_mod)
    return ret

# TODO _check_required_env_vars and _check_required_cmds have similar
# structures, maybe refactor to more generic
def _check_required_env_vars(metric_mod):
    UNSET = object()
    missing_env_vars = []
    required_env_vars = getattr(metric_mod, 'required_env_vars', {})
    for env_var, instr_text in required_env_vars.items():
        var_value = os.getenv(env_var)
        if not var_value:  # empty str is the same as unset
            print_err('env var "%s" is unset, but required by metric "%s". Instructions:\n\n%s\n'
                      % (env_var, metric_mod.__name__, instr_text))
            missing_env_vars.append(env_var)
    return missing_env_vars

def _check_required_cmds(metric_mod):
    missing_cmds = []
    required_cmds = getattr(metric_mod, 'required_cmds', {})
    for cmd, instr_text in required_cmds.items():
        cmd_path = shutil.which(cmd)
        if cmd_path is None:
            print_err('missing command "%s", required by metric "%s". Installation instructions:\n\n%s\n'
                      % (cmd, metric_mod.__name__, instr_text))
            missing_cmds.append(cmd)
    return missing_cmds


def collect_metrics(plist, repo_dir, metrics_dir, targets=None, metrics=None, dry_run=False, no_progress=False):
    "use local clones of repositories (from pull-repos) to gather data about projects"
    project_list = plist.project_list
    if targets:
        project_list = [proj for proj in project_list if (proj.name in targets or proj.name_slug in targets)]

    metric_mods = all_metric_mods = _get_all_metric_mods()
    if metrics:
        metric_mods = [m for m in metric_mods if m.__name__ in metrics]

    if not metric_mods:
        print_err('failed to collect data. no known metrics selected (available: %s)' % ', '.join([m.__name__ for m in all_metric_mods]))
        return

    project_repo_map = _get_project_repo_info_map(project_list, repo_dir)

    if not project_repo_map:
        print_err('failed to collect data on any projects')
        return

    res_fn_tmpl = 'apatite-metrics__{run_dt}__{newest_dt}__{oldest_dt}.jsonl'

    newest_dt = max([last_pulled for _, last_pulled in project_repo_map.values()])
    oldest_dt = min([last_pulled for _, last_pulled in project_repo_map.values()])

    res_fn = res_fn_tmpl.format(run_dt=datetime.datetime.utcnow().isoformat(),
                                # TODO: upgrade req to 3.6 and make isofromat('minutes'),
                                newest_dt=newest_dt.isoformat().rsplit(':', 1)[0],
                                oldest_dt=oldest_dt.isoformat().rsplit(':', 1)[0])

    print('collecting data for metrics: %s' % ', '.join([m.__name__ for m in metric_mods]))
    for metric_mod in metric_mods:
        with tqdm(total=len(project_repo_map), unit='run', disable=no_progress) as cur_metric_progress:
            for project, (target_repo_dir, last_pulled) in project_repo_map.items():
                cur_metric_progress.set_description(f'{metric_mod.__name__}: {project.name_slug}'.ljust(20))
                start_time = time.time()
                try:
                    res = metric_mod.collect(plist, project, target_repo_dir)
                except KeyboardInterrupt:
                    return
                except Exception as e:
                    print_err('%s metric failed on project %s with exception %r. (run with APATITE_DEBUG=1 to debug.)'
                              % (metric_mod.__name__, project.name_slug, e))
                    if os.getenv('APATITE_DEBUG'):
                        import pdb;pdb.post_mortem()
                    continue
                entry = {'project': project.name_slug,
                         'metric_name': metric_mod.__name__,
                         'result': res,
                         'pull_date': last_pulled.isoformat(),
                         '~ms': round((time.time() - start_time) * 1000, 2)}

                res_json = json.dumps(entry, sort_keys=True)
                if not dry_run:
                    with open(os.path.join(metrics_dir, res_fn), 'a') as f:
                        f.write(res_json + '\n')
                cur_metric_progress.update()
    return


def show_recent_metrics(metrics_dir):
    "shows the most recent metrics collection"
    metrics_files = sorted(iter_find_files(metrics_dir, '*.jsonl'), reverse=True)
    if not metrics_files:
        print_err('no recent metrics found at %s' % metrics_dir)
        return
    metrics_file = metrics_files[0]
    print('#  ' + os.path.basename(metrics_file) + '\n')
    for line in open(metrics_file):
        try:
            print(line, end='')
        except IOError:
            break
    return


def show_exportable_metrics(plist, earliest, metrics_dir, metrics=None):
    "show a list of metric names available for export-metrics --cols and --cols-file flags"
    return export_metrics(plist, earliest, metrics_dir, metrics, _show_exportable=True)


def export_metrics(plist, earliest, metrics_dir, metrics=None, output_path=None, output_format=None, _show_exportable=False):
    "export a csv with metrics collated from previous collect-metrics runs"
    metric_mods = all_metric_mods = _get_all_metric_mods(check_reqs=False)
    if metrics:
        metric_mods = [m for m in metric_mods if m.__name__ in metrics]
    if not metric_mods:
        print_err('failed to collect data. no known metrics selected (available: %s)' % ', '.join([m.__name__ for m in all_metric_mods]))
        return

    metrics_map = {(m.__name__, p.name_slug): None for m in all_metric_mods for p in plist.project_list}

    metrics_files = iter_find_files(metrics_dir, '*.jsonl')
    earliest_text = earliest.isoformat()
    files_to_search = []
    for metric_file in metrics_files:
        metric_base_fn = os.path.basename(os.path.splitext(metric_file)[0])
        _, run_dt_text, newest_dt_text, oldest_dt_text = metric_base_fn.split('__')
        if newest_dt_text < earliest_text:
            print('skipping', metric_file)
            continue
        files_to_search.append(metric_file)

        with open(metric_file) as f:
            # TODO: possible optimization when searching for a
            # specific project/metric. search for the project name
            # slug and metric name in the part of the line before the
            # result begins (the jsonl keys are carefully chosen to
            # sort nicely)
            for line_data in JSONLIterator(f):
                metric_name, proj_slug = line_data['metric_name'], line_data['project']
                try:
                    cur_data = metrics_map[metric_name, proj_slug]
                except KeyError:
                    # not a tracked project/metric
                    continue
                if cur_data is None or cur_data['pull_date'] < line_data['pull_date']:
                    metrics_map[metric_name, proj_slug] = line_data

    possible_paths = IndexedSet()
    for (metric_name, proj_slug), data in metrics_map.items():
        if data is None:
            continue
        def _visit(path, key, value):
            if not isinstance(value, (list, dict)):
                possible_paths.add((metric_name,) + path + (key,))
            return True
        remap(data['result'], visit=_visit)

    # TODO: deal with missing metrics
    # TODO: output csv or something

    ''' --cols 'license.total,evcs.*, sloc.TOTAL_*  --cols-file

    if col.endswith('*'):
       pop the segment with the star, fetch up until that point, then fetch/flatten everything underneath
    '''
    possible_paths = sorted(possible_paths)
    path_texts = ['.'.join('%s' % s for s in path) for path in possible_paths]

    from pprint import pprint
    if _show_exportable:
        print('\n'.join(path_texts))
        print('Showing %s exportable columns.' % len(possible_paths))
        return

    # for each project, output project_name, ...cols..., pull_date
    cols = path_texts
    all_proj_dicts = []
    for project in plist.project_list:
        cur_proj_dict = {'name': project.name_slug}
        for col in cols:
            metric_mod_name, glom_path = col.split('.', 1)
            cur_result_dict = (metrics_map[metric_mod_name, project.name_slug] or {'result': {}})['result']
            cur_proj_dict[col] = glom.glom(cur_result_dict, glom_path, default='')
            cur_proj_dict[col] = cur_proj_dict[col] if cur_proj_dict[col] is not None else ''
        all_proj_dicts.append(cur_proj_dict)

    all_cols = ['name'] + cols  # TODO: + ['pull_date'] (oldest of all the collated metrics or?

    with open('apatite_export.csv', 'w') as f:
        w = csv.DictWriter(f, all_cols)
        w.writeheader()
        for proj_dict in all_proj_dicts:
            w.writerow(proj_dict)

    print('exported %s columns for %s projects across %s metrics (%s)'
          % (len(all_cols), len(all_proj_dicts), len(metric_mods),
             ', '.join(sorted(m.__name__ for m in metric_mods))))

    return


class MetricsCollector(object):
    # maybe someday
    def __init__(self, project_list):
        self.project_processor = ProjectProcessor(self._collect_one, project_list)


def show_missing_tags():
    pass

# TODO: eventually going to need a command to migrate names in result files

"""
Read-only operations follow
"""

def print_version():
    'print the apatite version and exit'
    print('apatite version %s' % __version__)
    sys.exit(0)


"""
End subcommand handlers

Begin middlewares
"""

@face_middleware(provides=['plist', 'pdir', 'pfile'], optional=True)
def mw_ensure_project_listing(next_, file):
    file_path = file or 'projects.yaml'
    file_abs_path = os.path.abspath(file_path)

    if not os.path.exists(file_abs_path):
        raise APACLIError('Project listing not found: %s' % file_abs_path, 2)
    pdir = os.path.dirname(file_abs_path)
    plist = ProjectList.from_path(file_abs_path)
    return next_(plist=plist, pdir=pdir, pfile=file_abs_path)


@face_middleware(provides=['work_dir', 'repo_dir', 'metrics_dir'], optional=True)
def mw_ensure_work_dir(next_):
    work_dir = os.path.expanduser('~/.apatite/')
    repo_dir = work_dir + 'repos/'
    mkdir_p(repo_dir)
    open(repo_dir + '/.apatite_repo_dir', 'w').close()
    metrics_dir = work_dir + 'metrics/'
    mkdir_p(metrics_dir)
    return next_(work_dir=work_dir, repo_dir=repo_dir, metrics_dir=metrics_dir)


@face_middleware
def mw_exit_handler(next_):
    status = 55  # should always be set to something else
    try:
        try:
            status = next_() or 0
        except APACLIError:
            raise
        except APAError as ppe:
            raise APACLIError(ppe.args[0])
    except KeyboardInterrupt:
        status = 130
        print('')
    except EOFError:
        status = 1
        print('')
    except APACLIError as ppce:
        if ppce.args:
            print('Error: ' + '; '.join([str(a) for a in ppce.args]))
        status = ppce.exit_code

    sys.exit(status)
    return




import itertools
import os
import re

from boltons.dictutils import OMD
from boltons.iterutils import first
# from boltons.fileutils import iter_find_files  # vendored to include dirs

from apatite.utils import run_cap

RG_CMD = 'rg'
VERSION_CMD = 'vermin'

required_cmds = {RG_CMD: 'install from https://github.com/BurntSushi/ripgrep/releases (put on path, chmod a+x)',
                 VERSION_CMD: 'install version 0.6.0 of vermin from: https://pypi.org/project/vermin/#description'}

# probably going to have to manually code this up in here, and then
# modify the tagsonomy posthoc once things start looking coherent

def _join_patt_any_order(patt1, patt2):
    return '({0}.*{1})|({1}.*{0})'.format(patt1, patt2)


IMPORT_MAP = {'server': {'server_framework': ['django',
                                              'flask',
                                              'zope',
                                              'web2py',
                                              'pylons',
                                              'pyramid',
                                              'bottle'],},
              'desktop': {'gui_framework': ['qt',
                                            'gtk',
                                            'wx',
                                            'kivy',
                                            'pygame']},
              None: {'concurrency': ['gevent',
                                     'twisted',
                                     'tornado',
                                     'concurrent\\.futures',
                                     'asyncio',
                                     ]}}  # i also checked for trio, but got a false positive in freecad


FREEZERS = ['pyinstaller', 'cx_Freeze', 'py2exe', 'py2app', 'pynsist']
CONTAINER_FILES = ['snapcraft.yaml', 'appimage', 'flatpak', 'Dockerfile']


def collect(plist, project, repo_dir):
    ret = {}
    ret['type'] = arch_type = _get_arch_type(plist, project)
    # proc_res = run_cap([RG_CMD])
    # raise SystemExit()

    if arch_type == 'unknown':
        print(arch_type, project.name)

    ret['dep'] = dep_map = {}
    if arch_type not in IMPORT_MAP:
        return ret

    _all_arches = IMPORT_MAP[None].items()
    for col, dep_names in itertools.chain(IMPORT_MAP[arch_type].items(), _all_arches):
        dep_res_map = {}
        for dep_name in dep_names:
            search_output = search_files(_join_patt_any_order('import', dep_name), '*.py', repo_dir)
            dep_res_map[dep_name] = search_output.splitlines()
        top_dep, top_dep_res = sorted(dep_res_map.items(), key=lambda x: len(x[1]))[-1]
        if top_dep_res:
            dep_map[col] = top_dep

    ret['pkg'] = _get_pkg_info(plist, project, repo_dir)

    ret['compat'] = _get_py_version(plist, project, repo_dir)
    '''
    # See comment at bottom of file
    if ret['type'] == 'server' and ret['pkg'].get('freezer'):
        print('server with freezer', project.name, ret['pkg']['freezer'])
    if ret['type'] == 'desktop' and ret['pkg'].get('container'):
        print('desktop with container', project.name, ret['pkg']['container'])
    '''
    return ret


def search_files(patt, file_patt, repo_dir):
    cmd = [RG_CMD]
    if file_patt:
        cmd.extend(['-g', file_patt])
    cmd.append(patt)
    proc_res = run_cap(cmd, cwd=repo_dir)
    return proc_res.stdout


# copied in from boltons to add include_dirs
import re
import fnmatch
basestring = (str, bytes)
def iter_find_files(directory, patterns, ignored=None, include_dirs=False):
    """Returns a generator that yields file paths under a *directory*,
    matching *patterns* using `glob`_ syntax (e.g., ``*.txt``). Also
    supports *ignored* patterns.

    Args:
        directory (str): Path that serves as the root of the
            search. Yielded paths will include this as a prefix.
        patterns (str or list): A single pattern or list of
            glob-formatted patterns to find under *directory*.
        ignored (str or list): A single pattern or list of
            glob-formatted patterns to ignore.

    For example, finding Python files in the current directory:

    >>> filenames = sorted(iter_find_files(_CUR_DIR, '*.py'))
    >>> os.path.basename(filenames[-1])
    'urlutils.py'

    Or, Python files while ignoring emacs lockfiles:

    >>> filenames = iter_find_files(_CUR_DIR, '*.py', ignored='.#*')

    .. _glob: https://en.wikipedia.org/wiki/Glob_%28programming%29

    """
    if isinstance(patterns, basestring):
        patterns = [patterns]
    pats_re = re.compile('|'.join([fnmatch.translate(p) for p in patterns]))

    if not ignored:
        ignored = []
    elif isinstance(ignored, basestring):
        ignored = [ignored]
    ign_re = re.compile('|'.join([fnmatch.translate(p) for p in ignored]))
    for root, dirs, files in os.walk(directory):
        if include_dirs:
            for basename in dirs:
                if pats_re.match(basename):
                    if ignored and ign_re.match(basename):
                        continue
                    filename = os.path.join(root, basename)
                    yield filename

        for basename in files:
            if pats_re.match(basename):
                if ignored and ign_re.match(basename):
                    continue
                filename = os.path.join(root, basename)
                yield filename
    return



def _get_arch_type(plist, project):
    '''type: desktop, server, mobile, browser
    install_targets: windows, mac, linux, android, ...
    web frameworks: django, flask, other  (tornado, zope, etc.)
    gui_frameworks: gtk, qt, qt4, qt5, qt45, tk, wx, pygame, kivy, beeware, other
    cli_frameworks: argparse, optparse, click, etc.
    container: docker, flatpak, snap, appimage

    Is docker a target? Is docker linux?

    desktop with install target linux is different than server, where
    linux is all but assumed. docker implies server.

    basic definitions:
    * desktop = mostly single-user/local software.
    * server = mostly multi-user software
    * mobile/browser = a little less unclear

    TODO: what to do about deluge, which is single-user, but also has
    a multi-user-accessible server mode? or jupyter which is
    single-user but has a server architecture? probably have to tie-break based on

    '''

    tags = [t for t in project._tags if t in plist.tag_registry]  # TODO: why is _tags underscored again?

    if 'server' in tags:
        return 'server'
    desktop_tags = [t for t in tags if 'desktop' in plist.tag_registry[t].tag_path]
    if desktop_tags:
        return 'desktop'
    mobile_tags = [t for t in tags if 'mobile' in plist.tag_registry[t].tag_path]
    if mobile_tags:
        return 'mobile'
    return 'unknown'


def _get_pkg_info(plist, project, repo_dir):
    # snap: search for snapcraft.yaml
    # appimage: find -iname "appimage" -type d
    # flatpak: find -iname "flatpak" -type d  # maybe exclude test dirs, e.g., what ansible has
    # docker: find -name "Dockerfile"
    ret = {}
    container_stacks = OMD()
    for path in iter_find_files(repo_dir, CONTAINER_FILES, include_dirs=True):
        container_stacks.add(os.path.splitext(os.path.basename(path))[0].lower(), path)
    #if container_stacks:
    #    print(container_stacks.todict())
    has_docker = bool(container_stacks.pop('dockerfile', None))
    container_stack = first(container_stacks.keys(), None) or ('docker' if has_docker else '')
    ret['container'] = container_stack

    # TODO: split into mac/windows/linux? for linux I'll need to look
    # at deb/rpm, and I'm not sure the best strategy there. rpm maybe
    # .spec files? might have to check inside as other tools
    # (pyinstaller) uses .spec, too.

    # freezers -> pyInstaller, cx_Freeze, py2exe, py2app, pynsist
    # (bbFreeze phased out, osnap/constructor not yet adopted, harder
    # to search for). conda and omnibus also not adopted.

    freezer_res_map = OMD()
    for freezer_name in FREEZERS:
        search_output = search_files(freezer_name, '*', repo_dir)
        if search_output:
            freezer_res_map.add(freezer_name, len(search_output.splitlines()))
    if freezer_res_map:
        top, top_res = sorted(freezer_res_map.items(), key=lambda x: x[1])[-1]
        ret['freezer'] = top

    return ret


def _get_py_version(plist, project, repo_dir):

    # -v flag required because it improves the correctness of detection
    # manual tests on the calibre package showed better results with -v
    cmd = [VERSION_CMD, '-v', repo_dir]
    proc_res = run_cap(cmd, cwd=repo_dir)
    # hardcoded strings to search output lines for
    MIN_VERSION_PATTERN = '^Minimum required versions:'
    INCOMP_VERSION_PATTERN = '^Incompatible versions:'

    lines = proc_res.stdout.splitlines()
    min_version_line = None
    incomp_version_line = None
    # with -v flag a lot of output is generated, the lines containing
    # version information will probably be at the bottom
    for line in lines[-5:]:
        if re.match(MIN_VERSION_PATTERN, line):
            min_version_line = line
        if re.match(INCOMP_VERSION_PATTERN, line):
            incomp_version_line = line

    # None indicates not compatible with that major version of python
    py_2_version = None
    py_3_version = None
    if min_version_line is not None:
        supp_versions = min_version_line[26:].split()
        for version in supp_versions:
            if version[0] == '2':
                py_2_version = version.strip(',')
            elif version[0] == '3':
                py_3_version = version.strip(',')

    return {'min_py2': py_2_version, 'min_py3': py_3_version}


''' Some interesting results that illustrate how server/desktop
packaging aren't really that disjoint. Some of the docker ones are
desktop applications for developers, or use docker for development
setup, but others are legit:

desktop with container ArchiveBox docker
server with freezer Deluge py2app
server with freezer SABnzbd pyinstaller
desktop with container ZeroNet docker
desktop with container cartoonify / Draw This. docker
desktop with container FreeCAD docker
desktop with container Meshroom docker
desktop with container OCRopus docker
server with freezer Octoprint py2exe
desktop with container Ranger docker
desktop with container PyMedusa docker
desktop with container Bitmessage docker
desktop with container Hangups snapcraft
desktop with container Mailpile docker
server with freezer MoinMoin py2exe
desktop with container TahoeLAFS docker
desktop with container InVesalius docker
desktop with container Manim docker
desktop with container Sage Math docker
desktop with container Mercurial docker
server with freezer Review Board py2exe
desktop with container Ansible docker
desktop with container pgcli docker
server with freezer StackStorm py2exe
desktop with container MITMproxy docker
server with freezer Pupy py2exe
server with freezer Spiderfoot py2exe
desktop with container Universal Radio Hacker (URH) docker
desktop with container Komodo Edit docker
desktop with container pipenv docker
desktop with container yum docker
server with freezer buildbot py2exe
desktop with container Meson docker
desktop with container Pants docker
desktop with container asciinema docker
desktop with container Home Assistant docker
'''




import re
from collections import Counter

import attr
from boltons.timeutils import isoparse
from boltons.dictutils import OMD

from apatite.utils import run_cap

required_cmds = {'git': 'install via your package manager'}

def collect(plist, project, repo_dir):
    vcs_name = project.clone_info[0]
    ret = {'vcs_name': vcs_name}
    if vcs_name == 'git':
        ret.update(get_git_info(repo_dir))
    return ret


@attr.s(cmp=False)
class Committer(object):
    names = attr.ib(default=attr.Factory(set))
    emails = attr.ib(default=attr.Factory(set))
    commit_count = attr.ib(default=0)

    def merged(self, other):
        ret = type(self)(names=set(self.names),
                         emails=set(self.emails),
                         commit_count=self.commit_count)
        ret.names.update(other.names)
        ret.emails.update(other.emails)
        ret.commit_count += other.commit_count
        return ret


class CommitterRegistry(object):
    """Built to handle the aliasing that occurs in git logs where the
    same person will appear with different names and emails, sometimes
    due to changes over time, sometimes due to committing from
    different environments, etc.

    Consolidates sets of names, emails, and commit counts down to
    unified identities the best we can.

    """
    def __init__(self):
        self._email_to_ident = {}
        self._name_to_ident = {}

    def register(self, name, email, count):
        norm_name = ' '.join(name.lower().split())
        norm_email = email.lower()  # TODO: maybe remove plus segments

        email_ident = self._email_to_ident.get(norm_email)
        name_ident = self._name_to_ident.get(norm_name)

        if name_ident == email_ident:
            ident = email_ident
        else:
            if name_ident is None:
                ident = email_ident
            elif email_ident is None:
                ident = name_ident
            else:
                ident = email_ident.merged(name_ident)
                for email in ident.emails:
                    self._email_to_ident[email] = ident
                for names in ident.names:
                    self._name_to_ident[name] = ident
        if ident is None:
            ident = Committer()
        ident.names.add(norm_name)
        self._name_to_ident[norm_name] = ident

        ident.emails.add(norm_email)
        self._email_to_ident[norm_email] = ident

        ident.commit_count += count
        return

    def get_committers(self):
        return sorted(set(self._email_to_ident.values()), key=lambda x: x.commit_count, reverse=True)


def _get_commit_dt(repo_dir, commit_hash, **kw):
    kw.setdefault('env', {})['TZ'] = 'UTC'
    kw['cwd'] = repo_dir
    proc_res = run_cap(['git', 'show', '-s', '--format=%cd', '--date=format-local:%Y-%m-%dT%H:%M:%S', commit_hash], **kw)
    date_text = proc_res.stdout.strip()
    return isoparse(date_text)


_git_committer_re = re.compile(r'^\s+(?P<commit_count>\d+)\s+(?P<name>.*)\s<(?P<email>[^>]*)>$', re.MULTILINE | re.UNICODE)

def get_git_info(repo_dir):
    ret = {}

    proc_res = run_cap(['git', 'rev-list', '--max-parents=0', 'HEAD'], cwd=repo_dir)
    first_commit_hashes = proc_res.stdout.strip().split()

    first_commit_dt = sorted([_get_commit_dt(repo_dir, fch) for fch in first_commit_hashes])[0]

    proc_res = run_cap(['git', 'rev-parse', 'HEAD'], cwd=repo_dir)
    latest_commit_hash = proc_res.stdout.strip()

    latest_commit_dt = _get_commit_dt(repo_dir, latest_commit_hash)

    ret['first_commit'] = first_commit_dt.isoformat()
    ret['latest_commit'] = latest_commit_dt.isoformat()

    proc_res = run_cap(['git', 'shortlog', '--summary', '--numbered', '--email'], cwd=repo_dir)

    committer_registry = CommitterRegistry()
    for match in _git_committer_re.finditer(proc_res.stdout):
        gdict = match.groupdict()
        gdict['commit_count'] = int(gdict['commit_count'])

        committer_registry.register(gdict['name'], gdict['email'], gdict['commit_count'])

    committers = committer_registry.get_committers()
    ret['commit_count'] = commit_count = sum([c.commit_count for c in committers])
    ret['committer_count'] = len(committers)  # redundant with committer_percent_dist.100

    # these will be stored as percentages, so keep it to two-digit precision max
    threshes = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.95, 0.99, 1.0]
    commit_thresh_map = {thresh: (commit_count * thresh) for thresh in threshes}

    sorted_committers = sorted([(c, c.commit_count) for c in committers],
                               reverse=True, key=lambda x: x[1])
    def _get_proportion_count(thresh_commit_count):
        _cur_commit_count = 0
        _cur_committer_count = 0
        for committer, committer_commit_count in sorted_committers:
            if _cur_commit_count > thresh_commit_count:
                break
            _cur_commit_count += committer_commit_count
            _cur_committer_count += 1
        return _cur_committer_count

    # how many developers' commits does it take to comprise XX% of the commits?
    committer_dist_map = {round(thresh * 100): _get_proportion_count(thresh_commit_count)
                          for thresh, thresh_commit_count in commit_thresh_map.items()}
    ret['committer_percent_dist'] = committer_dist_map
    ret['committer_top_5'] = [round(c / commit_count, 4) for _, c in sorted_committers][:5]
    ret['minor_committer_counts'] = {x: len([c for _, c in sorted_committers if c <= x])
                                     for x in range(1, 6)}

    '''
    # DEBUG
    print(first_commit_dt.isoformat(), latest_commit_dt.isoformat(), latest_commit_dt - first_commit_dt)
    from pprint import pprint
    pprint(committer_dist_map)
    pprint(ret['top_5'])
    pprint(ret)
    raise SystemExit  # quits after the first
    '''
    return ret
